"""Datastructures not in the standard library.

Implements: `RingList`, `merge_dicts`"""

from __future__ import annotations
from typing import Any


class RingList:
    def __init__(self, max_size: int):
        """Initialize a RingList with a maximum size."""

        assert max_size > 0, "a max_size of zero doesn't make any sense"
        self._max_size: int = max_size
        self._data: list[float] = [0 for _ in range(max_size)]
        self._current_index: int = -1  # -1 means empty

    def clear(self) -> None:
        """Removes all elements from the list."""

        self._current_index = -1

    def is_full(self) -> bool:
        """Returns True if the list is full."""

        return self._current_index >= (self._max_size - 1)

    def append(self, x: float) -> None:
        """Appends an element to the list."""
        self._current_index += 1
        bounded_current_index = self._current_index % self._max_size
        self._data[bounded_current_index] = x

    def get(self) -> list[float]:
        """Returns the list of elements."""
        # list is full
        if self._current_index >= (self._max_size - 1):
            bounded_current_index = self._current_index % self._max_size
            return (
                self._data[bounded_current_index + 1 : self._max_size + 1] +
                self._data[0 : bounded_current_index + 1]
            )

        # list is not empty but not full
        elif self._current_index >= 0:
            return self._data[0 : self._current_index + 1]

        # list is empty
        return []

    def sum(self) -> float:
        """Returns the max size of the list"""
        return sum(self.get())

    def get_max_size(self) -> int:
        return self._max_size

    def set_max_size(self, new_max_size: int) -> None:
        """Sets a new max size fo the list."""
        current_list = self.get()
        self._max_size = new_max_size
        self._data = [0] * new_max_size
        self._current_index = -1
        for item in current_list:
            self.append(item)

    def __str__(self) -> str:
        return str(self.get())


def merge_dicts(old_object: Any, new_object: Any) -> Any:
    """
    For a given dict, update it recursively from a new dict.
    It will not add any properties and assert that the types
    remain the same (or null). null->int or int->null is possible
    but not int->dict or list->int.

    example:
    ```python
    merge_dicts(
        old_object={"a": 3, "b": {"c": 50, "e": None}},
        new_object={"b": {"e": 80}},
    ) == {"a": 3, "b": {"c": 50, "e": 80}}
    ```
    """

    if old_object is None or new_object is None:
        return new_object

    # merging is only possible when both are dicts
    if (type(old_object) == dict) and (type(new_object) == dict):
        updated_dict = {}

        old_keys = set(old_object.keys())
        new_keys = set(new_object.keys())

        # recursively merge keys that are in both dicts
        for k in old_keys.intersection(new_keys):
            updated_dict[k] = merge_dicts(old_object[k], new_object[k])

        # simple add keys that are only in one of the dicts
        for k in new_keys.difference(old_keys):
            updated_dict[k] = new_object[k]
        for k in old_keys.difference(new_keys):
            updated_dict[k] = old_object[k]

        return updated_dict
    else:
        return new_object



"""Common system status related functions.

Implements: `get_cpu_usage`, `get_memory_usage`, `get_disk_space`,
`get_system_battery`, `get_last_boot_time`, `get_utc_offset`"""

from __future__ import annotations
from typing import Any, Optional
import time
import psutil
import datetime


def get_cpu_usage() -> list[float]:
    """Checks the CPU usage of the system.

    Returns:
        The CPU usage in percent for each core."""

    return psutil.cpu_percent(interval=1, percpu=True)


def get_memory_usage() -> float:
    """Checks the memory usage of the system.

    Returns:
        The memory usage in percent."""

    p = psutil.virtual_memory().percent
    assert isinstance(p, float)
    return p


def get_disk_space(path: str = "/") -> float:
    """Checks the disk space of a given path.

    Args:
        path: The path to check the disk space for.

    Returns:
        The available disk space in percent."""

    return psutil.disk_usage(path).percent


def get_system_battery() -> Optional[int]:
    """Checks the system battery.
    
    Returns:
        The battery state in percent if available, else None."""

    battery_state: Any = psutil.sensors_battery()  # type: ignore
    try:
        assert battery_state is not None
        percent = battery_state.percent
        assert isinstance(percent, int)
        assert 1 <= percent <= 100
        return percent
    except AssertionError:
        return None


def get_last_boot_time() -> datetime.datetime:
    """Checks the last boot time of the system."""

    return datetime.datetime.fromtimestamp(psutil.boot_time())


def get_utc_offset() -> float:
    """Returns the UTC offset of the system.

    ```python
    x = get_utc_offset()
    local time == utc time + x
    ```
    
    Credits to https://stackoverflow.com/a/35058476/8255842
    
    Returns:
        The UTC offset in hours."""

    return round((-time.timezone) / 3600, 3)
